# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import os
import thermosteam as tmo

__all__ = ('create_chemicals',)
    
def create_chemicals():
    """Create chemicals for the production of fatty alcohols."""
    filepath = os.path.dirname(__file__)
    chemical_data_path = os.path.join(filepath, 'chemicals.yaml') 
    chemical_data = tmo.ThermoData.from_yaml(chemical_data_path)
    return chemical_data.create_chemicals()


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""
from biosteam import System
import biosteam as bst
import thermosteam as tmo
from thermosteam import Stream
from biorefineries.cornstover._process_settings import price
from biorefineries.cornstover._chemicals import chemical_groups
from biorefineries.cornstover import units
import thermosteam.reaction as rxn
import numpy as np


__all__ = ('create_system',)

def find_split(IDs, flow0, flow1):
    flow0 = np.asarray(flow0)
    splits = flow0/(flow0 + np.asarray(flow1))
    chemicals = tmo.settings.get_chemicals()
    array = np.zeros(chemicals.size)
    for ID, split in zip(IDs, splits):
        if ID in chemical_groups:
            array[chemicals.get_index(chemical_groups[ID])] = split
        else:
            array[chemicals.index(ID)] = split
    return array

def create_system(ID='cornstover_sys'):
    System.maxiter = 400
    System.converge_method = 'Aitken'
    System.molar_tolerance = 0.01
    
    ### Streams ###    
    chemicals = bst.settings.get_chemicals()
    
    # feed flow
    dry_composition = chemicals.kwarray(
        dict(Glucan=0.3505, Xylan=0.1953, Lignin=0.1576,
             Ash=0.0493, Acetate=0.0181, Protein=0.0310,
             Extract=0.1465, Arabinan=0.0238, Galactan=0.0143,
             Mannan=0.0060, Sucrose=0.0077)
    )
    moisture_content = chemicals.kwarray(
        dict(Water=0.20)
    )
    netflow = 104167.0
    feedflow = netflow*(dry_composition*0.8 + moisture_content)
    
    cornstover = Stream('cornstover',
                        feedflow,
                        units='kg/hr',
                        price=price['Feedstock'])
    warm_process_water = Stream('warm_process_water',
                             T=368.15,
                             P=4.7*101325,
                             Water=140000,
                             units='kg/hr')
    ammonia_process_water = Stream('ammonia_process_water',
                             T=368.15,
                             P=4.7*101325,
                             Water=150310,
                             units='kg/hr')
    rectifier_bottoms_product = Stream('',
                                       T=100+273.15,
                                       P=101325,
                                       Ethanol=18,
                                       Water=36629,
                                       Furfural=72,
                                       HMF=100,
                                       units='kg/hr')
    sulfuric_acid = Stream('sulfuric_acid',
                           P=5.4*101325,
                           T=294.15,
                           Water=139,
                           SulfuricAcid=1842,
                           units='kg/hr',
                           price=price['Sulfuric acid'])
    steam = Stream('steam',
                   phase='g',
                   T=268+273.15,
                   P=13*101325,
                   Water=24534+3490,
                   units='kg/hr')
    ammonia = Stream('ammonia',
                     Ammonia=1051,
                     units='kg/hr',
                     phase='l',
                     price=price['Ammonia'])
    cellulase = Stream('cellulase',
                       units='kg/hr',
                       price=price['Enzyme'])
    
    ### Pretreatment system
    
    U101 = units.FeedStockHandling('U101', cornstover)
    U101.cost_items['System'].cost = 0
    T201 = units.SulfuricAcidTank('T201', sulfuric_acid)
    M201 = units.SulfuricAcidMixer('M201', (rectifier_bottoms_product, T201-0))
    M202 = bst.Mixer('M202', (M201-0, warm_process_water, U101-0))
    M203 = units.SteamMixer('M203', (M202-0, steam), P=5.5*101325)
    R201 = units.PretreatmentReactorSystem('R201', M203-0)
    P201 = units.BlowdownDischargePump('P201', R201-1)
    T202 = units.OligomerConversionTank('T202', P201-0)
    F201 = units.PretreatmentFlash('F201', T202-0, P=101325, Q=0)
    M204 = bst.Mixer('M204', (R201-0, F201-0))
    H201 = units.WasteVaporCondenser('H201', M204-0, T=99+273.15, V=0)
    M210 = units.AmmoniaMixer('M210', (ammonia, ammonia_process_water))
    M205 = bst.Mixer('M205', (F201-1, M210-0))
    T203 = units.AmmoniaAdditionTank('T203', M205-0)
    P209 = units.HydrolyzatePump('P209', T203-0)
    H301 = units.HydrolysateCooler('H301', P209-0, T=48+273.15)
    M301 = units.EnzymeHydrolysateMixer('M301', (H301-0, cellulase))
    
    def update_pretreatment_process_water():
        warm_process_water.imass['Water'] = 140000.0 * M202.ins[2].F_mass / 104175.33336
    
    sulfuric_acid_over_feed = sulfuric_acid.mol/cornstover.F_mass
    def update_sulfuric_acid_loading():
        # Also plant air
        F_mass_feed = cornstover.F_mass
        plant_air.mol[0] = 0.8 *F_mass_feed
        sulfuric_acid.mol[:] = sulfuric_acid_over_feed * F_mass_feed
    
    hydrolyzate = F201.outs[1]
    ammonia_over_hydrolyzate = ammonia.mol/310026.22446428984
    def update_ammonia_loading():
        ammonia.mol[:] = ammonia_over_hydrolyzate * hydrolyzate.F_mass
        ammonia_process_water.imass['Water'] = ammonia.imass['Ammonia'] * 150310 / 1051
    
    cooled_hydrolyzate = H301.outs[0]
    enzyme_over_cellulose = 20/1000 * 20 # (20 g enzyme / cellulose) / (50 g cellulase / 1 L enzyme)
    water_cellulase_mass = cellulase.imass['Water', 'Cellulase']
    water_cellulase_to_cellulase = np.array([0.95, 0.05])
    def update_cellulase_and_nutrient_loading():
        F_mass_cooled_hydrolyzate = cooled_hydrolyzate.F_mass
        cellulose = cornstover.imass['Glucan']
        # Note: An additional 10% is produced for the media glucose/sophorose mixture
        # Humbird (2011) pg. 37 
        water_cellulase_mass[:] = (enzyme_over_cellulose
                                   * water_cellulase_to_cellulase
                                   * cellulose * 1.1)
        DAP1.mol[:] = DAP1_over_hydrolyzate * F_mass_cooled_hydrolyzate
        DAP2.mol[:] = DAP2_over_hydrolyzate * F_mass_cooled_hydrolyzate
        CSL1.mol[:] = CSL1_over_hydrolyzate * F_mass_cooled_hydrolyzate
        CSL2.mol[:] = CSL2_over_hydrolyzate * F_mass_cooled_hydrolyzate
        
    pretreatment_sys = System('pretreatment_sys',
                   path=(U101, T201, M201, update_pretreatment_process_water, 
                         M202, M203, R201, P201, T202, F201, M204, H201, M210,
                         M205, T203, P209, T203, H301, M301))
    
    ### Fermentation system ###
    
    DAP1 = Stream('DAP1',
                    DAP=26,
                    units='kg/hr',
                    price=price['DAP'])
    DAP2 = Stream('DAP2',
                    DAP=116,
                    units='kg/hr',
                    price=price['DAP'])
    DAP_storage = units.DAPTank('DAP_storage', Stream('DAP_fresh'), outs='DAP')
    
    S301 = bst.ReversedSplitter('S301', DAP_storage-0, outs=(DAP1, DAP2))
    CSL1 = Stream('CSL1',
                    CSL=211,
                    units='kg/hr',
                    price=price['CSL'])
    CSL2 = Stream('CSL2',
                    CSL=948,
                    units='kg/hr',
                    price=price['CSL'])
    CSL_storage = units.CSLTank('CSL_storage', Stream('CSL_fresh'), outs='CSL')
    
    S302 = bst.ReversedSplitter('S302', CSL_storage-0, outs=(CSL1, CSL2))
    denaturant = Stream('denaturant',
                        Octane=230.69,
                        units='kg/hr',
                        price=price['Denaturant'])
    stripping_water = Stream('stripping_water',
                              Water=26836,
                              units='kg/hr')
    DAP1_over_hydrolyzate = DAP1.mol/451077.22446428984
    DAP2_over_hydrolyzate = DAP2.mol/451077.22446428984
    CSL1_over_hydrolyzate = CSL1.mol/451077.22446428984
    CSL2_over_hydrolyzate = CSL2.mol/451077.22446428984
    
    M302 = bst.Mixer('M302', (M301-0, None))
    R301 = units.SaccharificationAndCoFermentation('R301', (M302-0, CSL2, DAP2))
    M303 = bst.Mixer('M303', (R301-2, CSL1, DAP1))
    R302 = units.SeedTrain('R302', M303-0)
    T301 = units.SeedHoldTank('T301', R302-1)
    T301-0-1-M302
    
    fermentation_sys = System('fermentation_sys',
                   path=(update_cellulase_and_nutrient_loading,
                         M302, R301, M303, R302, T301),
                   recycle=M302-0)
    
    ### Ethanol purification ###
    
    M304 = bst.Mixer('M304', (R302-0, R301-0))
    T302 = units.BeerTank('T302')
    
    M401 = bst.Mixer('M401', (R301-1, None))
    M401-0-T302
    
    D401 = bst.VentScrubber('D401', (stripping_water, M304-0),
                            gas=('CO2', 'NH3', 'O2'))
    D401-1-1-M401
    
    # Heat up before beer column
    # Exchange heat with stillage
    H401 = bst.HXprocess('H401', (T302-0, None),
                         phase0='l', phase1='l', U=1.28)
    
    Ethanol_MW = chemicals.Ethanol.MW
    Water_MW = chemicals.Water.MW
    def Ethanol_molfrac(e):
        """Return ethanol mol fraction in a ethanol water mixture"""
        return e/Ethanol_MW / (e/Ethanol_MW + (1-e)/Water_MW)
    
    # Beer column
    xbot = Ethanol_molfrac(0.00001)
    ytop = Ethanol_molfrac(0.50)
    D402 = bst.BinaryDistillation('D402', H401-0, k=1.25, Rmin=0.3,
                           P=101325, y_top=ytop, x_bot=xbot,
                           LHK=('Ethanol', 'Water'))
    D402.tray_material = 'Stainless steel 304'
    D402.vessel_material = 'Stainless steel 304'
    D402.boiler.U = 1.85
    P401 = bst.Pump('P401', D402-1)
    P401-0-1-H401
    
    # Mix ethanol Recycle (Set-up)
    M402 = bst.Mixer('M402', (D402-0, None))
    
    ytop = Ethanol_molfrac(0.915)
    D403 = bst.BinaryDistillation('D403', M402-0, Rmin=0.3,
                           P=101325, y_top=ytop, x_bot=xbot,
                           k=1.25, LHK=('Ethanol', 'Water'))
    D403.tray_material = 'Stainless steel 304'
    D403.vessel_material = 'Stainless steel 304'
    D403.is_divided = True
    D403.boiler.U = 1.85
    P402 = bst.Pump('P402', D403-1)
    
    P402-0-0-M201
    
    # Superheat vapor for mol sieve
    H402 = bst.HXutility('H402', D403-0, T=115+273.15, V=1)
    
    # Molecular sieve
    U401 = bst.MolecularSieve('U401', H402-0,
                             split=(2165.14/13356.04, 1280.06/1383.85),
                             order=('Ethanol', 'Water'))
    
    U401-0-1-M402
    
    ethanol_recycle_sys = System('ethanol_recycle_sys',
                                 path=(M402, D403, H402, U401),
                                 recycle=M402-0)
    
    # Condense ethanol product
    H403 = bst.HXutility('H403', U401-1, V=0, T=350.)
    
    T701 = bst.StorageTank('T701', H403-0, tau=7*24,
                           vessel_type='Floating roof',
                           vessel_material='Carbon steel')
    P701 = bst.Pump('P701', T701-0)
    
    # Storage for gasoline
    T702 = bst.StorageTank('T702', denaturant, tau=7*24,
                           vessel_type='Floating roof',
                           vessel_material='Carbon steel')
    P702 = bst.Pump('P702', T702-0)
    
    # Mix in denaturant
    ethanol = Stream('ethanol', price=price['Ethanol'])
    M701 = bst.MixTank('M701', (P702-0, P701-0), outs=ethanol)
    M701.line = 'Mixer'
    M701.tau = 0.05
    
    def adjust_denaturant():
        denaturant.imol['Octane'] = 0.022*P701.outs[0].F_mass/114.232
    
    P401.BM = P402.BM = P701.BM = P702.BM = 3.1
    T701.BM = T702.BM = 1.7
    
    vent_stream = M304-0
    stripping_water_over_vent = stripping_water.mol / 21202.490455845436
    def update_stripping_water():
        stripping_water.mol[:] = stripping_water_over_vent * vent_stream.F_mass
    
    puresys = System('purification',
                     path=(M304,
                              update_stripping_water,
                              D401, 
                              M401, T302,
                              H401, D402,
                              H401, P401, H401,
                              ethanol_recycle_sys,
                              P402, H403, T701, P701,
                              adjust_denaturant,
                              T702, P702, M701))
    
    ### Lignin Separation
    recycled_water = tmo.Stream(Water=1,
                                T=47+273.15,
                                P=3.9*101325,
                                units='kg/hr')
    
    splits = [('Glucose', 19, 502),
              ('Xylose', 40, 1022),
              ('OtherSugars', 81, 2175),
              ('SugarOligomers', 60, 1552),
              ('OrganicSolubleSolids', 612, 15808),
              ('InorganicSolubleSolids', 97, 2513),
              ('Furfurals', 19, 513),
              ('OtherOrganics', 52, 1348),
              ('Glucan', 1230, 25),
              ('Xylan', 415, 8),
              ('OtherStructuralCarbohydrates', 94, 2),
              ('Lignin', 12226, 250),
              ('Protein', 3376, 69),
              ('CellMass', 925, 19),
              ('OtherInsolubleSolids', 4489, 92)]
    
    S401 = units.PressureFilter('S401', ('', recycled_water),
                                moisture_content=0.35,
                                split=find_split(*zip(*splits)))
    H401-1-0-S401
    
    ### Waste water treatment ###
    
    combustion = chemicals.get_combustion_reactions()
    def growth(reactant):
        f = chemicals.WWTsludge.MW / getattr(chemicals, reactant).MW
        return rxn.Reaction(f"{f}{reactant} -> WWTsludge", reactant, 1.)
        
    organic_groups = ['OtherSugars', 'SugarOligomers', 'OrganicSolubleSolids',
                      'Furfurals', 'OtherOrganics', 'Protein', 'CellMass']
    organics = list(sum([chemical_groups[i] for i in organic_groups],
                        ('Ethanol', 'AceticAcid', 'Xylose', 'Glucose')))
    organics.remove('WWTsludge')
    
    P_sludge = 0.05/0.91/chemicals.WWTsludge.MW
    MW = np.array([chemicals.CH4.MW, chemicals.CO2.MW])
    mass = np.array([0.51, 0.49])*MW
    mass /= mass.sum()
    mass *= 0.86/(0.91)
    P_ch4, P_co2 = mass/MW
    def anaerobic_rxn(reactant):
        MW = getattr(chemicals, reactant).MW
        return rxn.Reaction(f"{1/MW}{reactant} -> {P_ch4}CH4 + {P_co2}CO2 + {P_sludge}WWTsludge",
                            reactant, 0.91)
    
    # TODO: Revise this with Jeremy
    anaerobic_digestion = rxn.ParallelReaction([anaerobic_rxn(i) for i in organics] + 
                                               [rxn.Reaction("H2SO4 -> H2S + 2O2", 'H2SO4', 1.)])
    
    
    # Note, nitogenous species included here, but most of it removed in R601 digester
    # TODO: Add ammonium to reaction, make sure it can be a liquid, possibly add Henry's constant
    aerobic_digestion = rxn.ParallelReaction([i*0.74 + 0.22*growth(i.reactant)
                                              for i in combustion
                                              if (i.reactant in organics)])
    aerobic_digestion.X[:] = 0.96
    
    splits = [('Ethanol', 1, 15),
              ('Water', 27158, 356069),
              ('Glucose', 3, 42),
              ('Xylose', 7, 85),
              ('OtherSugars', 13, 175),
              ('SugarOligomers', 10, 130),
              ('OrganicSolubleSolids', 182, 2387),
              ('InorganicSolubleSolids', 8, 110),
              ('Ammonia', 48, 633),
              ('AceticAcid', 0, 5),
              ('Furfurals', 5, 70),
              ('OtherOrganics', 9, 113),
              ('Cellulose', 19, 6),
              ('Xylan', 6, 2),
              ('OtherStructuralCarbohydrates', 1, 0),
              ('Lignin', 186, 64),
              ('Protein', 51, 18),
              ('CellMass', 813, 280),
              ('OtherInsolubleSolids', 68, 23)]
    
    well_water = Stream('well_water', Water=1, T=15+273.15)
    M601 = bst.Mixer('M601', (S401-1, '', ''))
    H201-0-1-M601
    
    WWTC = units.WasteWaterSystemCost('WWTC', M601-0)
    R601 = units.AnaerobicDigestion('R601', (WWTC-0, well_water),
                                     reactions=anaerobic_digestion,
                                     sludge_split=find_split(*zip(*splits)))
    
    air = Stream('air_lagoon', O2=51061, N2=168162, phase='g', units='kg/hr')
    caustic = Stream('WWT_caustic', Water=2252, NaOH=2252,
                     units='kg/hr', price=price['Caustic']*0.5)
    # polymer = Stream('WWT polymer') # Empty in humbird report :-/
    
    M602 = bst.Mixer('M602', (R601-1, None))
    
    caustic_over_waste = caustic.mol / 2544300.6261793654
    air_over_waste = air.mol / 2544300.6261793654
    waste = M602-0
    def update_aerobic_input_streams():
        F_mass_waste = waste.F_mass
        caustic.mol[:] = F_mass_waste * caustic_over_waste
        air.mol[:] = F_mass_waste * air_over_waste
    
    R602 = units.AerobicDigestion('R602', (waste, air, caustic),
                                  outs=('evaporated_water', ''),
                                  reactions=aerobic_digestion)
    
    splits = [('Ethanol', 0, 1),
              ('Water', 381300, 2241169),
              ('Glucose', 0, 2),
              ('Xylose', 1, 3),
              ('OtherSugars', 1, 7),
              ('SugarOligomers', 1, 6),
              ('OrganicSolubleSolids', 79, 466),
              ('InorganicSolubleSolids', 4828, 28378),
              ('Ammonia', 3, 16),
              ('Furfurals', 0, 3),
              ('OtherOrganics', 1, 7),
              ('CarbonDioxide', 6, 38),
              ('O2', 3, 17),
              ('N2', 5, 32),
              ('Cellulose', 0, 194),
              ('Xylan', 0, 65),
              ('OtherStructuralCarbohydrates', 0, 15),
              ('Lignin', 0, 1925),
              ('Protein', 0, 90),
              ('CellMass', 0, 19778),
              ('OtherInsolubleSolids', 0, 707)]
    
    S601 = bst.Splitter('S601', R602-1, split=find_split(*zip(*splits)))
    
    S602 = bst.Splitter('S602', S601-1, split=0.96)
    
    M603 = bst.Mixer('M603', (S602-0, None))
    M603-0-1-M602
    
    M604 = bst.Mixer('M604', (R601-2, S602-1))
    
    centrifuge_species = ('Water', 'Glucose', 'Xylose', 'OtherSugars',
                          'SugarOligomers', 'OrganicSolubleSolids',
                          'InorganicSolubleSolids', 'Ammonia', 'Furfurals', 
                          'OtherOrganics', 'CO2', 'COxSOxNOxH2S', 'Cellulose',
                          'Xylan', 'OtherStructuralCarbohydrates', 'Lignin',
                          'Protein', 'CellMass', 'OtherInsolubleSolids')
    S623_flow = np.array([7708, 0, 0, 1, 1, 13, 75, 3, 0, 1, 1, 2, 25, 8, 2, 250, 52, 1523, 92])
    S616_flow = np.array([109098, 3, 6, 13, 9, 187, 1068, 46, 5, 8, 14, 31, 1, 0, 0, 13, 3, 80, 5])
    
    S603 = bst.Splitter('S603', M604-0, outs=('', 'sludge'),
                      split=find_split(centrifuge_species, S616_flow, S623_flow))
    S603-0-1-M603
    
    S604 = bst.Splitter('S604', S601-0, outs=('treated_water', 'waste_brine'),
                      split={'Water': 0.987})
    
    aerobic_digestion_sys = System('aerobic_digestion_sys',
                                   path=(M602, update_aerobic_input_streams, R602, S601, S602, M604, S603, M603),
                                   recycle=M602-0)
    
    ### Facilities
    
    M501 = bst.Mixer('M501', (S603-1, S401-0))
    BT = bst.facilities.BoilerTurbogenerator('BT',
                                             ins=(M501-0, R601-0, 
                                                  'boiler_makeup_water',
                                                  'natural_gas',
                                                  'lime',
                                                  'boilerchems'), 
                                             turbogenerator_efficiency=0.85)
    
    CWP = bst.facilities.ChilledWaterPackage('CWP')
    CT = bst.facilities.CoolingTower('CT')
    CT.outs[1].T = 273.15 + 28
    
    process_water_streams = (caustic,
                             stripping_water,
                             warm_process_water,
                             steam, BT-1, CT-1)
            
    makeup_water = Stream('makeup_water', price=price['Makeup water'])
    
    PWC = bst.facilities.ProcessWaterCenter('PWC',
                                            (S604-0, makeup_water),
                                            (),
                                            None,
                                            (BT-1, CT-1),
                                            process_water_streams)
    blowdown_mixer = bst.BlowdownMixer('blowdown_mixer', (BT-1, CT-1), 2**M601)
    
    CIP = Stream('CIP', Water=126, units='kg/hr')
    CIP_package = units.CIPpackage('CIP_package', CIP)
    plant_air = Stream('plant_air', N2=83333, units='kg/hr')
    ADP = bst.facilities.AirDistributionPackage('ADP', plant_air)
    fire_water = Stream('fire_water', Water=8343, units='kg/hr')
    FT = units.FireWaterTank('FT', fire_water)
    
    ### Complete system
    
    cornstover_sys = System(ID,
                            path=(pretreatment_sys, fermentation_sys,
                                  puresys, S401, M601, WWTC, R601,
                                  aerobic_digestion_sys, S604),
                            facilities=(M501, CWP, BT, CT,
                                        PWC, ADP, CIP_package, S301, S302,
                                        DAP_storage, CSL_storage, FT, blowdown_mixer),
                            facility_recycle=blowdown_mixer-0)
    
    return cornstover_sys



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import numpy as np
import biosteam as bst
from thermosteam.utils import get_instance
from . import units as units

__all__ = ('LAOsProcessSpecifications',
           'fermentation_products',
           'load_process_settings')

# %% Constant specifications

fermentation_products = ('Hexanol','Octanol', 'Decanol', 
                         'Dodecanol', 'Tetradecanol')

prices = [0.2378, 0.2756, 0.3171]
Ts = [412.189, 454.77, 508.991]
coef = np.polyfit(Ts, prices, 1)
calculate_steam_price_at_T = np.poly1d(coef)

# %% Overall process specifications

def load_process_settings():
    bst.process_tools.default_utilities()
    
    # CEPCI
    bst.CE = 567.5 # 2017
    
    # High pressure steam conditions is a process specification that
    # depends on dehydration reactor temperature
    high_pressure_steam = bst.HeatUtility.get_heating_agent('high_pressure_steam')
    high_pressure_steam.heat_transfer_efficiency = 0.95
    
    # About 10 degC higher than boiling point of tridecane
    medium_pressure_steam = bst.HeatUtility.get_heating_agent('medium_pressure_steam')
    medium_pressure_steam.heat_transfer_efficiency = 0.95
    medium_pressure_steam.T = 510
    medium_pressure_steam.P = medium_pressure_steam.chemicals.Water.Psat(510)
    medium_pressure_steam.regeneration_price = calculate_steam_price_at_T(510)
    
    # Low pressure steam, T is default
    low_pressure_steam = bst.HeatUtility.get_heating_agent('low_pressure_steam')
    low_pressure_steam.T = 412.189
    low_pressure_steam.P = 344738.0
    low_pressure_steam.regeneration_price = 0.2378
    low_pressure_steam.heat_transfer_efficiency = 0.95

class LAOsProcessSpecifications:
    """
    Create an LAOsProcessSpecifications object that handles overall
    process and technology specifications of the LAOs production system.
    
    """
    __slots__ = ('system',
                 'dehydration_reactor',
                 'TEA',
                 'fermentation_specification',
                 'dehydration_reactor_temperature',
                 'dehydration_reactor_conversion',
                 'LAOs_over_alcohol_arr',
                 'operating_days',
                 'plant_size',
    )
    
    def __init__(self, system, TEA,
                 plant_size=150000,
                 operating_days=350.4, 
                 fermentation_titer=40,
                 fermentation_productivity=1.0,
                 fermentation_yield=0.90,
                 dehydration_reactor_temperature=623.15,
                 dehydration_reactor_mass_fraction=0.2,
                 dehydration_reactor_conversion=0.9999,
        ):
        #: [System] Should simulate the complete process.
        self.system = system
        
        #: [TEA] Should include settings for the complete process.
        self.TEA = TEA
        
        fermentation_reactor = get_instance(system.units, units.FattyAlcoholBioreactor)
        
        #: [AdiabaticFixedbedGasReactor] Dehydration reactor.
        self.dehydration_reactor = get_instance(system.units, units.AdiabaticFixedbedGasReactor)
        
        #: [ReactorSpecification] Specifications for fementor.
        self.fermentation_specification = bst.process_tools.ReactorSpecification(
            reactor = fermentation_reactor, 
            reaction_name = 'fermentation_reaction',
            substrates = ('Glucose',),
            products = fermentation_products,
            titer = fermentation_titer, 
            productivity = fermentation_productivity,
            yield_ = fermentation_yield, 
            production = 100., # Temporary
        )
        
        MW_alcohols = np.array([i.MW for i in fermentation_reactor.chemicals[fermentation_products]]) 
        #: [1d array] Stoichiometric coefficients by weight of fatty alcohol to LAOs.
        self.LAOs_over_alcohol_arr = (MW_alcohols - 18.01528) / MW_alcohols
        
        self.dehydration_reactor_temperature = dehydration_reactor_temperature 
        self.operating_days = operating_days 
        self.plant_size = plant_size 
        
        #: [float] Mass fraction of alcohols entering dehydration reaction.
        #: Setting this parameter does nothing directly, but it is used
        #: elsewhere in the process simulation.
        self.dehydration_reactor_mass_fraction = dehydration_reactor_mass_fraction
        
        #: [float]Conversion of alcohols in dehydration reaction.
        self.dehydration_reactor_conversion = dehydration_reactor_conversion
    
    def run_specifications(self, *args, **kwargs):
        """Load specifications and simulate system."""
        self.load_specifications(*args, **kwargs)
        self.system.simulate()
      
    def load_specifications(self,
                           operating_days=None,
                           dehydration_reactor_temperature=None,
                           fermentation_titer=None,
                           fermentation_productivity=None,
                           fermentation_yield=None,
                           plant_size=None,
                           dehydration_reactor_mass_fraction=None,
                           dehydration_reactor_conversion=None):
        """Load all specifications without simulating the system."""
        self.load_operating_days(operating_days or self.operating_days)
        self.load_dehydration_reactor_temperature(dehydration_reactor_temperature or self.dehydration_reactor_temperature)
        fspec = self.fermentation_specification
        fspec.load_yield(fermentation_yield or self.fermentation_yield)
        fspec.load_titer(fermentation_titer or self.fermentation_titer)
        fspec.load_productivity(fermentation_productivity or self.fermentation_productivity)
        self.load_dehydration_reactor_conversion(dehydration_reactor_conversion or self.dehydration_reactor_conversion)
        self.load_dehydration_reactor_mass_fraction(dehydration_reactor_mass_fraction or self.dehydration_reactor_mass_fraction)
        self.load_plant_size(plant_size or self.plant_size)
        
    @property
    def LAOs_over_alcohol(self):
        """[float] LAOs per fatty alcohol ratio in g LAOs / g fatty alcohol."""
        fermentation_specificaiton = self.fermentation_specification
        reaction = fermentation_specificaiton.reaction
        if reaction.basis != 'wt': reaction = reaction.copy('wt')
        products = reaction.istoichiometry[fermentation_specificaiton.products]
        products /= products.sum()
        return (products * self.LAOs_over_alcohol_arr).sum()
    
    @property
    def fermentation_yield(self):
        return self.fermentation_specification.yield_
    @fermentation_yield.setter
    def fermentation_yield(self, yield_):
        self.fermentation_specification.yield_ = yield_
    
    @property
    def fermentation_titer(self):
        return self.fermentation_specification.titer
    @fermentation_titer.setter
    def fermentation_titer(self, titer):
        self.fermentation_specification.titer = titer
    
    @property
    def fermentation_productivity(self):
        return self.fermentation_specification.productivity
    @fermentation_productivity.setter
    def fermentation_productivity(self, productivity):
        self.fermentation_specification.productivity = productivity
    
    @property
    def dehydration_reactor_mass_fraction(self):
        """
        [float] Mass fraction of alcohols entering dehydration reaction.
        Setting this parameter does nothing directly, but it is used
        elsewhere in the process simulation.
        """
        return self.dehydration_reactor.dehydration_reactor_mass_fraction
    @dehydration_reactor_mass_fraction.setter
    def dehydration_reactor_mass_fraction(self, dehydration_reactor_mass_fraction):
        self.dehydration_reactor.dehydration_reactor_mass_fraction = dehydration_reactor_mass_fraction
    
    def load_operating_days(self, operating_days):
        """
        Load operating days specification.
        
        Parameters
        ----------
        operating_days : float
            Number of operating days per year for TEA.
            
        """
        self.operating_days = self.TEA.operating_days = operating_days
    
    def load_plant_size(self, plant_size):
        """
        Load plant size specification.
        
        plant_size : float 
            Plant size in ton product / yr of LAOs.
        
        Notes
        -----
        The net flow rate of feed to bioreactor is adjusted to achieve this
        specification.
        
        """
        self.plant_size = plant_size
        self.fermentation_specification.load_production(
            plant_size / self.LAOs_over_alcohol * 907.18474 / 24. / self.operating_days
        )
    
    def load_dehydration_reactor_conversion(self, dehydration_reactor_conversion):
        self.dehydration_reactor.reaction.X[:] = self.dehydration_reactor_conversion = dehydration_reactor_conversion
    
    def load_dehydration_reactor_temperature(self, dehydration_reactor_temperature):
        """
        Load dehydration reactor temperature specification.
        
        Parameters
        ----------
        dehydration_reactor_temperature : float
            Alcohol dehydration fixed bed reactor temperature in Kelvin.
            
        Notes
        -----
        The temperature of high pressure steam and its regeneration price
        are adjusted so that its 10 K higher that reactor operating temperature.
        
        """
        self.dehydration_reactor.T = T_reactor = dehydration_reactor_temperature
        
        # Higher pressure steam is required to heatup reactor feed
        high_pressure_steam = bst.HeatUtility.get_heating_agent('high_pressure_steam')
        high_pressure_steam.T = T_steam = T_reactor + 10
        high_pressure_steam.regeneration_price =  calculate_steam_price_at_T(T_reactor)
        high_pressure_steam.P = high_pressure_steam.chemicals.Water.Psat(T_steam)
        
    def load_dehydration_reactor_mass_fraction(self, dehydration_reactor_mass_fraction):
        """
        Load dehydration reactor mass fraction specification.

        Parameters
        ----------
        dehydration_reactor_mass_fraction : float
            Mass fraction of alcohols entering dehydration reaction.
        
        Notes
        -----
        Setting this parameter does nothing directly, but it is used
        elsewhere in the process simulation.

        """
        self.dehydration_reactor_mass_fraction = dehydration_reactor_mass_fraction


# -*- coding: utf-8 -*-
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Part of the BioSTEAM project. Under the UIUC open-source license.
# See github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from math import pi

__all__ = ('AdiabaticFixedbedGasReactor', 'cylinder_diameter_from_volume',)

def cylinder_diameter_from_volume(volume, length_to_diameter):
    return (4. * length_to_diameter * volume / pi)**(1./3.)

def cylinder_volume_from_diameter(diameter, length_to_diameter):
    return pi * (diameter / 2) ** 2 * diameter / length_to_diameter

class AdiabaticFixedbedGasReactor(bst.design_tools.PressureVessel, bst.Unit):
    _N_ins = _N_outs = 1
    _N_heat_utilities = 0
    _units = {**bst.design_tools.PressureVessel._units,
              'Volume': 'ft^3',
              'Catalyst weight': 'kg'}
    _BM = {**bst.design_tools.PressureVessel._BM,
           'Catalyst': 1.0}
    
    def __init__(self, ID="", ins=None, outs=(), thermo=None, *,
                 reaction, catalyst_density, catalyst_price, WHSV,
                 P, length_to_diameter=3.0,
                 vessel_material='Stainless steel 304',
                 vessel_type='Vertical'):
        bst.Unit.__init__(self, ID, ins, outs, thermo)
        self.length_to_diameter = length_to_diameter #: Length to diameter ratio
        self.vessel_material = vessel_material # Vessel material
        self.vessel_type = vessel_type # 'Horizontal' or 'Vertical'
        self.reaction = reaction
        self.catalyst_density = catalyst_density # In kg/m^3
        self.catalyst_price = catalyst_price # In USD/kg
        self.WHSV = WHSV #: Residence time in kg/hr feed / kg catalyst
        self.P = P    
    
    def _default_vessel_type(self):
        return 'Vertical'
    
    def _setup(self):
        effluent, = self.outs
        effluent.P = self.P
        effluent.phase = 'g'
    
    def _run(self):
        feed, = self.ins
        assert feed.phase == 'g', 'feed phase must be a gas'
        effluent, = self.outs
        effluent.copy_like(feed)
        self.reaction.adiabatic_reaction(effluent)
        
    def _design(self):
        effluent, = self.outs
        length_to_diameter = self.length_to_diameter
        P = effluent.get_property('P', 'psi')
        results = self.design_results
        catalyst = effluent.F_mass / self.WHSV
        results['Catalyst weight'] = catalyst
        results['Volume'] = volume = catalyst / self.catalyst_density * 35.3147 # to ft3
        D = cylinder_diameter_from_volume(volume, length_to_diameter)
        L = length_to_diameter * D
        results.update(self._vessel_design(P, D, L))
        
    def _cost(self):
        D = self.design_results
        self.purchase_costs.update(
            self._vessel_purchase_cost(D['Weight'], D['Diameter'], D['Length'])
        )
        self.purchase_costs['Catalyst'] = self.catalyst_price * D['Catalyst weight']



# -*- coding: utf-8 -*-
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Part of the BioSTEAM project. Under the UIUC open-source license.
# See github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst

__all__ = ('SurgeTank',)

class SurgeTank(bst.StorageTank):
    _N_outs = 2
    _stream_link_options = bst.Unit._stream_link_options
    
    def _run(self):
        vent, effluent = self.outs
        effluent.copy_like(self.ins[0])
        vent.copy_thermal_condition(effluent)
        vent.copy_flow(effluent, 'N2', remove=True)
        
    def _design(self):
        design_results = self.design_results
        design_results['Residence time'] = tau = self.tau
        design_results['Total volume'] = tau * self.outs[1].F_vol / self.V_wf


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

import os
from doctest import testfile
from biorefineries import (
    sugarcane as sc,
    lipidcane as lc,
    cornstover as cs,
    LAOs as laos,
)
get_readme = lambda module: os.path.join(os.path.dirname(module.__file__), 'README.rst')
testfile(get_readme(sc))
testfile(get_readme(lc))
testfile(get_readme(cs))
testfile(get_readme(laos))


# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from thermosteam import functional as fn
import thermosteam as tmo

__all__ = ('create_chemicals',)

def create_chemicals():
    ### Define common chemicals ###
    
    Biodiesel = tmo.Chemical('Biodiesel',
                             search_ID='Methyl oleate')
    lipidcane_chemicals = tmo.Chemicals(
        ['Water', 'Methanol', 'Ethanol', 'Glycerol',
         'Glucose', 'Sucrose', 'H3PO4', 'P4O10', 'CO2',
         'Octane', 'O2', Biodiesel, 'CH4'])
    
    (Water, Methanol, Ethanol,
     Glycerol, Glucose, Sucrose,
     H3PO4, P4O10, CO2, Octane, O2, Biodiesel, CH4) = lipidcane_chemicals
    
    O2.at_state(phase='g')
    CH4.at_state(phase='g')
    CO2.at_state(phase='g')
    H3PO4.at_state(phase='s')
    P4O10.at_state(phase='s')
    Glucose.at_state(phase='s')
    Sucrose.at_state(phase='s')
    
    ### Define new chemicals ###
    
    def create_new_chemical(ID, phase='s', **constants):
        solid = tmo.Chemical.blank(ID, phase=phase, **constants)
        lipidcane_chemicals.append(solid)
        return solid
    
    Ash = create_new_chemical('Ash', MW=1.)
    Cellulose = create_new_chemical('Cellulose',
                                    formula="C6H10O5", # Glucose monomer minus water
                                    Hf=-975708.8)
    Hemicellulose = create_new_chemical('Hemicellulose',
                                        formula="C5H8O5", # Xylose monomer minus water
                                        Hf=-761906.4)
    Flocculant = create_new_chemical('Flocculant',
                                     MW=1.)
    Lignin = create_new_chemical('Lignin',
                                 formula='C8H8O3', # Vainillin
                                 Hf=-452909.632)
    Solids = create_new_chemical('Solids', MW=1.)
    DryYeast = create_new_chemical('DryYeast', MW=1., CAS='Yeast')
    CaO = create_new_chemical('CaO', formula='CaO')
    HCl = create_new_chemical('HCl', formula='HCl')
    NaOH = create_new_chemical('NaOH', formula='NaOH')
    NaOCH3 = create_new_chemical('NaOCH3', formula='NaOCH3')
    
    Lipid = create_new_chemical(
        'Lipid',
        phase = 'l',
        formula = 'C57H104O6',
        Hf = -2193.7e3
    )
    Lipid.Dortmund.set_group_counts_by_name({'CH3':3, 'CH2':41, 'CH':1, 'CH=CH':3, 'CH2COO':3})
    
    ### Fill missing properties ###
    
    # Assume properties are similar for trioleate and tripalmitin
    Tripalmitin = tmo.Chemical('Tripalmitin').at_state(phase='l', copy=True)
    Lipid.copy_models_from(Tripalmitin, ['V', 'sigma', 'kappa', 'Cn'])
    
    # Assume a constant volume for lipid
    lipid_molar_volume = fn.rho_to_V(rho=900, MW=Lipid.MW)
    Lipid.V.add_model(lipid_molar_volume)
    
    # Insolubles occupy a significant volume
    insoluble_solids = (Ash, Cellulose, Hemicellulose,
                        Flocculant, Lignin, Solids, DryYeast, P4O10)
    
    # Solubles don't occupy much volume
    soluble_solids = (CaO, HCl, NaOH, H3PO4, Glucose, Sucrose) 
    
    for chemical in insoluble_solids:
        V = fn.rho_to_V(rho=1540, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    for chemical in soluble_solids:
        V = fn.rho_to_V(rho=1e5, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    # Assume sodium methoxide has some of the same properities as methanol
    LiquidMethanol = Methanol.at_state(phase='l', copy=True)
    NaOCH3.copy_models_from(LiquidMethanol, ['V', 'sigma', 'kappa', 'Cn'])
    
    # Add constant models for molar heat capacity of solids
    Ash.Cn.add_model(0.09 * 4.184 * Ash.MW) 
    CaO.Cn.add_model(1.02388 * CaO.MW) 
    Cellulose.Cn.add_model(1.364 * Cellulose.MW) 
    Hemicellulose.Cn.add_model(1.364 * Hemicellulose.MW)
    Flocculant.Cn.add_model(4.184 * Flocculant.MW)
    Lignin.Cn.add_model(1.364 * Lignin.MW)
    Solids.Cn.add_model(1.100 * Solids.MW)
    
    for chemical in lipidcane_chemicals: chemical.default()
    
    lipidcane_chemicals.compile()
    lipidcane_chemicals.set_synonym('Water', 'H2O')
    return lipidcane_chemicals

